//! Space-filling curves and other related functionality.

#![allow(dead_code)]

pub mod hilbert {
    const N: u32 = 1 << 16;

    /// Convert (x,y) to hilbert curve index.
    ///
    /// x: The x coordinate.  Must be no greater than 2^16-1.
    /// y: The y coordinate.  Must be no greater than 2^16-1.
    ///
    /// Returns the hilbert curve index corresponding to the (x,y) coordinates given.
    pub fn encode(x: u32, y: u32, n: u32) -> u32 {
        assert!(x < N);
        assert!(y < N);

        let (mut x, mut y) = (x, y);
        let mut d = 0;
        let mut s = n >> 1;
        while s > 0 {
            let rx = if (x & s) > 0 { 1 } else { 0 };
            let ry = if (y & s) > 0 { 1 } else { 0 };
            d += s * s * ((3 * rx) ^ ry);
            (x, y) = hilbert_rotate(s, rx, ry, x, y);

            s >>= 1
        }

        d
    }

    /// Convert hilbert curve index to (x,y).
    ///
    /// d: The hilbert curve index.
    ///
    /// Returns the (x, y) coords at the given index.
    pub fn decode(d: u32, n: u32) -> (u32, u32) {
        let (mut x, mut y) = (0, 0);
        let mut s = 1;
        let mut t = d;
        while s < n {
            let rx = 1 & (t >> 1);
            let ry = 1 & (t ^ rx);
            (x, y) = hilbert_rotate(s, rx, ry, x, y);
            x += s * rx;
            y += s * ry;
            t >>= 2;

            s <<= 1;
        }

        (x, y)
    }

    //------------
    // Utilities.

    fn hilbert_rotate(n: u32, rx: u32, ry: u32, x: u32, y: u32) -> (u32, u32) {
        if ry == 0 {
            if rx == 1 {
                ((n - 1).wrapping_sub(y), (n - 1).wrapping_sub(x))
            } else {
                (y, x)
            }
        } else {
            (x, y)
        }
    }
}

pub mod morton {
    const N: u32 = 1 << 16;

    /// Convert (x,y) to morton curve index.
    ///
    /// x: The x coordinate.  Should be no greater than 2^16-1.
    /// y: The y coordinate.  Should be no greater than 2^16-1.
    ///
    /// Returns the morton curve index corresponding to the (x,y) coordinates given.
    pub fn encode(x: u32, y: u32) -> u32 {
        debug_assert!(x < N);
        debug_assert!(y < N);

        part_1_by_1(x) | (part_1_by_1(y) << 1)
    }

    /// Convert morton curve index to (x,y).
    ///
    /// i: The morton curve index.
    ///
    /// Returns the (x, y) coords at the given index.
    pub fn decode(i: u32) -> (u32, u32) {
        (compact_1_by_1(i), compact_1_by_1(i >> 1))
    }

    //------------
    // Utilities.

    #[inline(always)]
    fn part_1_by_1(mut x: u32) -> u32 {
        x &= 0x0000ffff;
        x = (x ^ (x << 8)) & 0x00ff00ff;
        x = (x ^ (x << 4)) & 0x0f0f0f0f;
        x = (x ^ (x << 2)) & 0x33333333;
        x = (x ^ (x << 1)) & 0x55555555;
        x
    }

    #[inline(always)]
    fn compact_1_by_1(mut x: u32) -> u32 {
        x &= 0x55555555;
        x = (x ^ (x >> 1)) & 0x33333333;
        x = (x ^ (x >> 2)) & 0x0f0f0f0f;
        x = (x ^ (x >> 4)) & 0x00ff00ff;
        x = (x ^ (x >> 8)) & 0x0000ffff;
        x
    }
}

/// Yields coordinates in outward spiral, but incorporating a Hilbert
/// curve at the smaller scales.
pub mod hilbert_spiral {
    /// Convert from hilbert-spiral index to (x,y).
    ///
    /// Note: this returns both negative and positive coordinates.
    /// It starts at 0,0 and spirals outwards.
    ///
    /// i: The hilbert-spiral index.
    /// hilbert_size: the size of the hulbert blocks on a side.  Will be
    ///               rounded down to the nearest power of two.
    ///
    /// Returns the (x, y) coords at the given index.
    pub fn decode(i: u32, hilbert_size: u32) -> (i32, i32) {
        assert!(hilbert_size > 0);
        let hilbert_size = 1 << (31 - u32::leading_zeros(hilbert_size));

        let hilbert_cells = hilbert_size * hilbert_size;

        let hilbert_i = i % hilbert_cells;
        let spiral_i = i / hilbert_cells;

        let (mut sx, mut sy, section) = decode_spiral(spiral_i);
        sx = (sx * hilbert_size as i32) - (hilbert_size / 2) as i32;
        sy = (sy * hilbert_size as i32) - (hilbert_size / 2) as i32;

        let (hx, hy) = {
            let (hx, hy) = super::hilbert::decode(hilbert_i, hilbert_size);
            let a = hilbert_size - 1;
            match section {
                0 => (hy, hx),
                1 => (a - hx, a - hy),
                2 => (a - hy, a - hx),
                3 => (hx, hy),
                _ => unreachable!(),
            }
        };

        (sx + hx as i32, sy + hy as i32)
    }

    pub fn decode_spiral(i: u32) -> (i32, i32, u32) {
        if i == 0 {
            return (0, 0, 3);
        }

        // 0 = first ring outside of center, 1 = second, and so on.
        let ring = (((i as f64).sqrt() - 1.0) / 2.0) as u32;

        // The size of the ring along one side.
        let size = 1 + ((ring + 1) * 2);

        let n = i - ((size - 2) * (size - 2)); // The zero-indexed cell of the ring.
        let arm = n / (size - 1); // The arm of the ring.
        let arm_n = n % (size - 1); // The index within the arm of the ring.

        // The two coordinates.  They just need to be flipped around depending on the arm.
        let radius = ring as i32 + 1;
        let d = -(size as i32 / 2) + 1 + arm_n as i32;

        match arm {
            0 => (radius, d, 0),
            1 => (-d, radius, if arm_n == (size - 2) { 2 } else { 1 }),
            2 => (-radius, -d, 2),
            3 => (d, -radius, 3),
            _ => unreachable!(),
        }
    }
}

//-------------------------------------------------------------

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn hilbert_reversible() {
        let i = 0x4c8587a2;
        let (x, y) = hilbert::decode(i);
        let i2 = hilbert::encode(x, y);

        assert_eq!(i, i2);
    }

    #[test]
    fn morton_reversible() {
        let i = 0x4c8587a2;
        let (x, y) = morton::decode(i);
        let i2 = morton::encode(x, y);

        assert_eq!(i, i2);
    }
}
